import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import arttoys.core.*; 
import java.util.Random; 
import oscP5.*; 
import netP5.*; 
import processing.sound.*; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class processing_gbloink extends PApplet {





IAutomatonToy toy;
Uids _ids = new Uids();
Random rnd = new Random();
IBus mainBus = new BasicBus();

ColorTool colorTool = new ColorTool(this);

NoteCalculator nc = new NoteCalculator();
IFreqStrategy fs = new ScaleBasedFreqStrategy(nc);  

ArrayList<IObservingController> instruments = new ArrayList<IObservingController>();

public GbloinkWorld makeGbloink() {
  nc.setCurrent("minor");
  Oscillator[] oscs = { new Pulse(this), new SinOsc(this), new TriOsc(this) };
  for (int i=0;i<3;i++) {
    //instruments.add( new ObInGbloink2ArtToys2("localhost", 9004, "/channel"+i, i, fs, mainBus ) );
      instruments.add( new BasicSynth(this,oscs[i],i,fs,mainBus) );
  }

  return new GbloinkWorld(new Ball(0,color(255,0,0),width/2 + random(100)-100,height/2+random(200)-100,2,2, mainBus),
                          new Ball(1,color(0,255,0),width/2 + random(100)-100,height/2+random(200)-100,2,2, mainBus),
                          new Ball(2,color(0,0,255),width/2 + random(100)-100,height/2+random(200)-100,2,2, mainBus),
                          colorTool
                          
  );
}

public void reset() {
  mainBus.reset();
  toy = makeGbloink();
  toy.reset();
  toy.sizeInSetup();  
}

int GLOBAL_WIDE = 1000;
int GLOBAL_HIGH = 700;

public void settings() {
  size(GLOBAL_WIDE,GLOBAL_HIGH);
}

public void setup() {
  reset();
  stroke(color(200,200,30));
}
 

public void draw() {
  mainBus.reset();
  toy.nextStep();  
  toy.draw();
  toy.postToBus();

  
  for (IObservingController in : instruments) {
    //println(in.getClass().getName());
    in.scanBus();
  }

}

public void mouseClicked() { toy.struck(mouseX,mouseY); }

public void keyPressed() {
  if (key==' ') { reset(); return; }
  toy.keyPressed(key); 
  nc.keyPressedUpdate(key);
}

public void mousePressed() { toy.mousePressed(mouseX, mouseY); }
public void mouseDragged() { toy.mouseDragged(mouseX, mouseY); }
public void mouseReleased() { toy.mouseReleased(); }
class StickyBall extends Ball {
  boolean following = false;

  StickyBall(int c, int col, float x, float y, float dx, float dy, IBus bus) { super(c,col,x,y,dx,dy,bus); }
  
  public void interact(Iterable<IMover> balls ) {
    if (following) { 
      setDX(other.getDX());
      setDY(other.getDY());      
      x=x+getDX();
      y=y+getDY();
    } else {
      x=x+getDX();
      y=y+getDY();

      if (outside(x,400)) { invDX(); }
      if (outside(y,400)) { invDY(); }
      
      for (IMover b2 : balls) {
        if (!(b2==this)) {
          if (PVector.sub(new PVector(b2.getX(),b2.getY()),new PVector(getX(),getY())).mag() < rad) {                      
            if (!((StickyBall)b2).following) {
              following = true;
              other = (Ball)b2;
            }
          }
        }
      }

    } 
    
  }
  
  
  public void draw() {
    fill(colour);
    ellipse(x,y,10,10);
    line(x,y,width/2,height/2);
  }  
}



class StickyBalls extends BaseControlAutomaton implements IAutomatonToy {
  
  ArrayList<IMover> balls;

  int no_balls=10;
  PVector centre;
  PVector target;
  int counter;

  StickyBalls(ColorTool ct, IBus bus) {
    centre = new PVector(200,200);  
  
    balls = new ArrayList<IMover>();
    RandomColorGenerator rc = ct.randomColorGenerator(false);
    for (int i=0;i<no_balls;i++) {
      balls.add(new StickyBall(i,rc.next(),random(300)+100,random(300)+100,random(5)+5,random(5)+5, bus));
    }

  }

  public void sizeInSetup() { size(400,400); }
  
  public void nextStep() {
    for (IMover b : balls) {
      b.interact(balls);
    }    
  }
  
  public void draw() {
    background(0);
    for (IMover b : balls) { 
      b.draw();
    }    
  }
  
  public void mousePressed(int mouseX, int mouseY) {
    target = new PVector(mouseX,mouseY);
    
    for (int i=0;i<no_balls;i++) {
      Ball b = (Ball)balls.get(i);
      b.head_for(target);
    } 
  }
  public void mouseReleased() { }

  public void struck(int x, int y) { }
  public void mouseDragged(int mouseX, int mouseY) { }
  public void reset() {
    // TODO something
  } 
  public void keyPressed(int k) { }
  
  public String diagnostic() { return "StickyBalls"; }
  
  public void postToBus() {
    // TODO something
  }
}


class BasicBlock extends BaseBlock {

  BasicBlock(int x, int y) {
    makeUid(_ids);
    
    setWidth(50);
    setHeight(50);

    setX(x);
    setY(y);    
  }  
  
  public void draw() {
    pushStyle();
    fill(200,200,200,100);
    stroke(255,255,255,255);
    rect(getX(),getY(),getWidth(),getHeight());
    popStyle();
  } 
  
}
PFont titleFont, descriptionFont;


class Ball extends BaseMover implements IObservable, IMover {
  
  PVector centre;
  int colour;
  Ball other;
  int rad;
  boolean bounced=false;
  
  Ball(int ch, int col, float x, float y, float dx, float dy, IBus bus) {
    setChannel(ch);
    this.x = x;
    this.y = y;
    this.rad = 10;   
    setDX(dx);
    setDY(dy);
    this.colour = col;
    setBus(bus);
  }

  public boolean outside(float n, int lim) {
    return ( (n < 0) || (n > lim) );
  }  

  public boolean hit(Ball b) {
    return close(b.getX(), b.getY(), x, y, b.rad + rad); 
  }

  public boolean hit(int ox, int oy) {
   return close(x,y,ox,oy,rad*2);
  }  
  
  public void interact(Iterable<IActor> blocks, Iterable<IMover> balls ) { 
    if (bounced) {
      bounced = false;
      return;
    }
    
    int tx=(int)(x+getDX());
    int ty=(int)(y+getDY());

    if (outside(tx,width)) { invDX(); bounced=true; }
    if (outside(ty,height)) { invDY(); bounced=true; }    
  
    for (IMover b : balls) {
      boolean flag1=false, flag2=false;
      if (!(b == this)) {
         if (b.hit(tx,getY())) {
           invDX();
           bounced=true;
           flag1=true;
         }
         if (b.hit(getX(),ty)) {
           invDY();
           bounced=true;
           flag2=true;
         }
         if (flag1 && (getDX() == b.getDX())) { setDX(-getDX()); }
         if (flag2 && (getDY() == b.getDY())) { setDY(-getDY()); }
      }      
    }  
  

    for (IActor b : blocks) {
      if (b.hit((int)(tx+getDX()),getY())) { 
        invDX(); 
        bounced=true;
      }
      if (b.hit(getX(),(int)(ty+getDY()))) {
        invDY();
        bounced=true;
      }
    } 

    x = tx;
    y = ty;
  }

  public String toString() { 
    return "Ball " + getChannel() + " : " + x + ", " + y;  
  }
  
  public void interact(Iterable<IMover> balls ) {
  }


  public void draw() {
    fill(colour);
    ellipse(x,y,rad*2,rad*2);
  }

  public void head_for(PVector target) {
    PVector vel = PVector.sub(target,new PVector(x,y));
    vel.normalize();
    vel.mult(5);
    setDX(vel.x);
    setDY(vel.y);
  }
  
  
  IBus innerBus;
  int channel;
  public void postToBus() { 
    if (bounced) {
      IMessage m = new FloatMessage( map(getX(),0,width,0,1), map(getY(),0,height,1,0), 0, 0, 0, 0 );    
      innerBus.put(getChannel(),m);

      m = new BangMessage();
      innerBus.put(getChannel(),m);
    }
  }
  public void setChannel(int c) { channel = c; }
  public int  getChannel() { return channel; }
  public void setBus(IBus b) { innerBus = b; }
  public IBus getBus() { return innerBus; }


  public String diagnostic() { return toString(); }  
}






class GbloinkBlock extends BasicBlock {
  int c;
  
  GbloinkBlock(int x, int y, RandomColorGenerator randCols) {
    super(x,y);
    
    setWidth((int)(10+Math.random() * 90));
    setHeight((int)(10+Math.random() * 90));

    setX(x-getWidth()/2);
    setY(y-getHeight()/2);
    c = makeTransparent(randCols.next(),200);    
  }    
  
  public void draw() {
    pushStyle();
    fill(c);
    stroke(255,255,255,255);
    rect(getX(),getY(),getWidth(),getHeight());
    popStyle();
  }
}

class GbloinkWorld extends BaseControlAutomaton implements IAutomatonToy, IBlockWorld {
  ArrayList<IMover> balls;
  BaseBlockWorld blocks; 
  int wide=GLOBAL_WIDE;
  int high=GLOBAL_HIGH;
   
  ColorTool ct;
  
  GbloinkWorld(IMover b1, IMover b2, IMover b3, ColorTool ct) {
    balls = new ArrayList<IMover>();
    balls.add(b1);
    balls.add(b2);
    balls.add(b3);
    this.ct = ct;
  
    titleFont = loadFont("BlippoBT-Black-64.vlw");
    descriptionFont = loadFont("Dialog.plain-32.vlw");
  }

  public void reset() {    
    blocks = new BaseBlockWorld();
    for (int i=0;i<wide;i=i+18) {
      addBlock(new GbloinkBlock(i,(int)(50+Math.random()*100),colorTool.randomColorGenerator(true)));
      addBlock(new GbloinkBlock(i,(int)(high-250+Math.random()*100),colorTool.randomColorGenerator(true)));
    }
  }
  
  public void nextStep() {
    for (IMover b : balls) {
      b.interact(blocks.itBlocks(),new IteratorCollection<IMover>(balls.iterator()));
    }
    
  }
  
  public void draw() { 
    background(0,0,66);
    blocks.draw();
    for (IMover b : balls) { b.draw(); } 
    fill(240,87,222);
    textSize(64);
    textFont(titleFont);
    text("Gbloink!",15,60);
    textSize(32);
    fill(69,219,105);
    textFont(descriptionFont);
    text("Synaesmedia since 1997. See gbloink.com and synaesmedia.net",15,height-30);
  }
  
  public void sizeInSetup() { 
    //size(wide,high);  
  }

  public void addBlock(IActor block) { blocks.addBlock(block); }
  public boolean blockSelected() { return blocks.blockSelected(); }
  public IActor selectedBlock() throws NoSelectedBlockException { return blocks.selectedBlock(); }
  public Iterable<IActor> itBlocks() { return blocks.itBlocks(); }

  public void postToBus() {
    for (IMover b: balls) { ((IObservable)b).postToBus(); }     
  } 

  public String diagnostic() { return "Gbloink! World with " + balls.size() + " balls"; } 

  public void mousePressed(int x, int y) { struck(x,y); }
  public void struck(int x, int y) {
     blocks.mousePressed(x,y);
     if (!(blocks.blockSelected())) {
       //addBlock(new GbloinkBlock(x,y,colorTool.randomColorGenerator(true))); 
     }
  }
  public void mouseReleased() { blocks.mouseReleased(); }  
  public void mouseDragged(int mouseX, int mouseY) { blocks.mouseDragged(mouseX, mouseY); }

  public void keyPressed(int k) { }
}

class ObInGbloink2ArtToys extends BaseObservingOSCInstrument {
    ObInGbloink2ArtToys(String ip, int port, String path, int chan, IFreqStrategy fs, IBus bus) {
      super(ip,port,path,chan,fs,bus);
    }
  
    public OscMessage makeMessage(int bang, float[] xs) {
      float tone = makeCorrectedFreq(xs[0]);
      float filter =  map(xs[0],0,1,0,1000);
      return mFact.make(bang, tone, filter, xs[2],xs[3],xs[4]);
    }
}

class ObInGbloink2ArtToys2 extends BaseObservingOSCInstrument {
    ObInGbloink2ArtToys2(String ip, int port, String path, int chan, IFreqStrategy fs, IBus bus) {
      super(ip,port,path,chan,fs,bus);
    }
  
    public OscMessage makeMessage(int bang, float[] xs) {
      float tone = makeCorrectedFreq(xs[1]);
      float filter =  map(xs[0],0,1,0,1000);
      return mFact.make(bang, tone, filter, xs[2],xs[3],xs[4]);
    }
}



OscP5 oscP5 = new OscP5(this,9001); // one, instantiated once 


interface IObservingOSCInstrument extends IObservingController, IMusicObserver {
  public void setIP(String ip, int port);
  public void setPath(String path);
  public OscP5 getOscP5();
  public NetAddress getRemoteLocation();
}


class OscMessageFactory {
  String path;
  OscMessageFactory(String p) { path = p; }
  
  public OscMessage make(String... params) {
      OscMessage msg = new OscMessage(path);
      for (int i=0; i<params.length;i++) {
        msg.add(params[i]);
      }
      return msg;
  }
  
  public OscMessage make(float... params) {
      OscMessage msg = new OscMessage(path);
      for (int i=0; i<params.length;i++) {
        msg.add(params[i]);
      }
      return msg;
  }
  
} 


abstract class BaseObservingOSCInstrument implements IObservingOSCInstrument {
  NetAddress myRemoteLocation;
  String path;
  IBus innerBus;
  IFreqStrategy fs;
  OscMessageFactory mFact;
  int channel;
 
  BaseObservingOSCInstrument(String ip, int port, String path, int chan, IFreqStrategy fs, IBus bus) {
    setIP(ip,port);
    setPath(path);  
    this.fs = fs;      
    channel = chan;
    setBus(bus);
  }
  
  
  public void setIP(String s, int port) {
    myRemoteLocation = new NetAddress(s,port);
  }

  public void setPath(String p) { 
    path=p;
    mFact = new OscMessageFactory(path); 
  }

  public NetAddress getRemoteLocation() { return myRemoteLocation; }
  public OscP5 getOscP5() { return oscP5; }

  public float makeCorrectedFreq(float y) { return fs.corrected(fs.rawFreq(y)); }

  public void setBus(IBus bus) { innerBus = bus; }
  public IBus getBus() { return innerBus; }
  public void postToBus() { }

  public IFreqStrategy getFreqStrategy() { return fs; }
  public void setFreqStrategy(IFreqStrategy ifs) { fs = ifs; }

  public void scanBus() {
    int bang = 0;  
    for (IMessage m : innerBus.getBangs(channel)) { bang=1; }
   
    for (IMessage m : innerBus.getFloats(channel)) {
      float[] xs = ((IFloatMessage)m).getFloats();
      OscMessage om = makeMessage(bang, xs);                                                                 
      oscP5.send(om , myRemoteLocation);
      om.print();
      
    }      
  }
   
  public abstract OscMessage makeMessage(int bang, float[] xs);  

}



Pulse pulse;

interface IObservingInternalSynth extends IObservingController, IMusicObserver {
  
}


class BasicSynth implements IObservingInternalSynth {
  IBus innerBus;
  IFreqStrategy fs;
  int channel;
  
  Oscillator osc;
  Env env; 
  Reverb reverb;
  Delay delay;
  
  float attackTime = 0.001f;
  float sustainTime = 0.004f;
  float sustainLevel = 0.3f;
  float releaseTime = 0.2f;
  
  BasicSynth(PApplet app, Oscillator o, int c, IFreqStrategy afs, IBus abus) {
    channel=c;
    fs = afs;
    innerBus = abus;

    osc = o; //new Pulse(app);
    osc.amp(0.3f);
    osc.play();  
    
    env = new Env(app);
    
    delay = new Delay(app);
    delay.feedback(0.3f);
    delay.process(osc, 0.25f);
    
    reverb = new Reverb(app);
    //reverb.process(osc);
    
    reverb.room(0.25f);
    reverb.damp(0.8f);
    reverb.wet(0.3f);
    
    
  }
  public void setBus(IBus bus) { innerBus = bus; }
  public IBus getBus() { return innerBus; }
  public void postToBus() { }
  public IFreqStrategy getFreqStrategy() { return fs; }
  public void setFreqStrategy(IFreqStrategy ifs) { fs = ifs; }
  public float makeCorrectedFreq(float y) { return fs.corrected(fs.rawFreq(y)); }
  public void scanBus() {
    int bang = 0;  
    for (IMessage m : innerBus.getBangs(channel)) { 
      env.play(osc,attackTime, sustainTime, sustainLevel, releaseTime);
    }
   
    for (IMessage m : innerBus.getFloats(channel)) {
      float[] xs = ((IFloatMessage)m).getFloats();
      osc.play(makeCorrectedFreq(xs[1]), 0.5f);
    }      
  }  
}

//Geometry
public boolean approx(float a, float b, float marg) {
    return ((a > b-marg) && (a < b+marg));
}

public boolean close(float x1, float y1, float x2, float y2, float dist) {
    float xdiff = x1-x2;
    float ydiff = y1-y2;
    return (sqrt(xdiff*xdiff+ydiff*ydiff) <= dist);
}

public int makeTransparent(int c, int a) {
   return color(red(c),green(c),blue(c),a); 
}
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "processing_gbloink" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
